# -*- coding: utf-8 -*-
"""
Provides support for template (engine) libraries.
A template library may provide additional functionality, like:

  * `filter`_ functions
  * `test`_ functions
  * `global`_ functions
  * `extension`_ classes (template directives, tags)

This additional functionality can then be used in templates.

.. _filter::    http://jinja.pocoo.org/docs/dev/templates/#filters
.. _test::      http://jinja.pocoo.org/docs/dev/templates/#tests
.. _global::    http://jinja.pocoo.org/docs/dev/api/#the-global-namespace
.. _extension:: http://jinja.pocoo.org/docs/dev/templates/#extensions


TEMPLATE LIBRARY EXAMPLE
--------------------------------

.. code-block:: python

    from jinja2_library import Library
    import jinja2.ext

    # -- LIBRARY:
    this_library  = Library()
    register = this_library.make_register_decorator()

    # -- LIBRARY CONTENTS:
    @register.filter
    def my_filter1(arg):
        return transform(arg)

    @register.filter(name="filter2", aliases=["my.filter2"])
    def my_filter2(arg):
        return transform(arg)

    # -- USE LIBRARY WITH JINJA:
    class LibraryExtension(jinja2.ext.Extension):
        # -- NOTE: Add this library functionality to the Jinja2 environment.
        # HINT: Simplifies its usage in Jinja2 template environment.
        def __init__(self, environment):
            super(LibraryExtension, self).__init__(environment)
            this_library.add_to_environment(environment)

.. note::

    The provided mechanism is similar to a `Django template library`_
    (``django.template.Library``). But the Library and the "register" decorator
    functionality is split up into two classes (for a cleaner, simpler design).

    .. _Django template library: https://docs.djangoproject.com/en/3.2/howto/custom-template-tags/
"""

from __future__ import absolute_import
import jinja2.ext
import six

# -----------------------------------------------------------------------------
# UTILITY FUNCTIONS:
# -----------------------------------------------------------------------------
def make_scoped_name(scope, name):
    scoped_name = name
    if scope:
        scoped_name = "%s.%s" % (scope, name)
    return scoped_name


def remove_scope_from_name(scope, name):
    if scope:
        prefix = "%s." % scope
        if name.startswith(prefix):
            name = name[len(prefix):]
    return name


def dict_assign_scoped(data, scoped_name, value):
    current_dict = data
    name_parts = scoped_name.split(".")
    last_part = name_parts[-1]

    # -- STEP: Select or create storage scope (as dict).
    for name_part in name_parts[:-1]:
        new_dict = current_dict.get(name_part, None)
        if new_dict is None:
            new_dict = current_dict[name_part] = {}
        current_dict = new_dict

    # -- STORAGE-SCOPE FOUND:
    current_dict[last_part] = value


def _ensure_environment_extensions_attribute_exists(environment):
    # -- PREVENT: Environment.extensions problem in constructor.
    _extensions = getattr(environment, "extensions", None)
    if _extensions is None:
        environment.extensions = {}


def add_library_to_environment(library, environment):
    """Add a template library (or similar) to the Jinja2 environment.

    A functional template library contains:

    * filters
    * tests
    * globals
    * extensions (extra tags; derived from: :class:`jinja2.ext.extension.Extension`)
    """
    # assert isinstance(environment, jinja2.environment.Environment)
    environment.filters.update(library.filters)
    environment.tests.update(library.tests)
    # -- SUPPORT: scoped/dotted names for globals
    # REQUIRES: Split up into multiple dicts
    for name, global_func in library.globals.items():
        if "." in name:
            dict_assign_scoped(environment.globals, name, global_func)
        else:
            environment.globals[name] = global_func

    _ensure_environment_extensions_attribute_exists(environment)
    for extension in library.extensions:
        environment.add_extension(extension)


# -----------------------------------------------------------------------------
# CLASS: Library
# -----------------------------------------------------------------------------
class Library(object):
    """Provides a template library (functionality).
    A template library provides additional functionality that can be used
    in templates, like:

    * Jinja2 filters (filter-functions)
    * Jinja2 tests (predicate-functions)
    * Jinja2 globals (global-functions)
    * extensions (tags, directives)

    The library object itself provides a bundle or modules of this extended
    functionality.

    Library Scopes
    --------------------

    The library can be assigned a :attr:`scope` that provides a namespace
    for its extensions.

    .. code:: python

        # -- FILE: jinja_ext_foo.py
        # TEMPLATE LIBRARY: foo (extension: "jinja_ext_foo:LibraryExtension")
        from jinja2_library import Library
        import jinja2.ext

        # -- TEMPLATE LIBRARY (bare bones):
        # HINT: Use Library() to use it without scope/namespace.
        this_library = Library(scope="foo")
        register = this_library.make_register_decorator()

        # -- LIBRARY CONTENTS:
        @register.filter
        def hello(arg):
            return "Hello %s" % arg

        @register.filter(name="hello2, aliases=["ciao"])
        def hello2(arg):
            return "Hello2 %s" % arg

        class LibraryExtension(jinja2.ext.Extension):
            def __init__(self, environment):
                super(LibraryExtension, self).__init__(environment)
                this_library.add_to_environment(environment)

    This scope prefix (as dotted name) must be used in a template
    to use the functional extensions of this library.

    .. code:: jinja

        # -- TEMPLATE EXAMPLE:
        {{"Alice"|foo.hello}}
        {{"Alice"|foo.hello2}}
        {{"Alice"|foo.ciao}}
    """

    def __init__(self, scope=None):
        self.scope = scope
        self.filters = {}
        self.tests = {}
        self.globals = {}
        self.extensions = []

    def make_scoped_name(self, name):
        return make_scoped_name(self.scope, name)

    def _add_to_data(self, data, func, name=None, aliases=None, aliases_scope=None):
        """Core functionality to add a filter, test or global."""
        if name is None:
            name = func.__name__
        scoped_name = self.make_scoped_name(name)
        data[scoped_name] = func
        if aliases:
            if aliases_scope is None:
                aliases_scope = self.scope
            for name in aliases:
                scoped_name = make_scoped_name(aliases_scope, name)
                data[scoped_name] = func
        return self

    def add_filter(self, func, name=None, aliases=None, aliases_scope=None):
        return self._add_to_data(self.filters, func, name=name,
                                 aliases=aliases,
                                 aliases_scope=aliases_scope)

    def add_test(self, func, name=None, aliases=None, aliases_scope=None):
        return self._add_to_data(self.tests, func, name=name,
                                 aliases=aliases,
                                 aliases_scope=aliases_scope)

    def add_global(self, func, name=None, aliases=None, aliases_scope=None):
        return self._add_to_data(self.globals, func, name=name,
                                 aliases=aliases,
                                 aliases_scope=aliases_scope)

    def add_extension(self, extension_class):
        if extension_class not in self.extensions:
            self.extensions.append(extension_class)
        return self

    def make_register_decorator(self, aliases_scope=None):
        return _RegisterToLibraryDecorator(self, aliases_scope)

    def add_to_environment(self, environment):
        add_library_to_environment(self, environment)

    # -- EXPERIMENT:
    # def make_extension_class(self, class_name=None):
    #     """Make a jinja2.ext.Extension class from this library object."""
    #     # -- HINT: Use this Library object as blueprint for Extension class.
    #     import jinja2.ext
    #     this_library = self
    #     class ExtensionClass(jinja2.ext.Extension):
    #         def __init__(self, environment):
    #             jinja2.ext.Extension.__init__(self, environment)
    #             this_library.add_to_environment(environment)
    #     ExtensionClass.__name__ = class_name or "Extension"
    #     return ExtensionClass


class _RegisterToLibraryDecorator(object):
    """Provides decorator functions to register filters, tests, globals and
    extensions (tags) to a library.

    Objects of this class are normally created by the library to which it
    registers to.

    .. code-block:: python

        from jinja2_library import Library
        this_library = Library()
        register = this_library.make_register_decorator()

        @register.filter
        def my_filter1(value, *args, **kwargs): pass

        @register.filter(name="hello", aliases=["hello2"])
        def my_filter2(value, *args): pass

        @register.filter("hello")
        def my_filter3(value, *args): pass

        # -- ALTERNATIVE:
        def my_filter4(value, *args): pass
        register.filter(my_filter4, name="filter4")
    """
    def __init__(self, lib, aliases_scope=None):
        self._lib = lib
        self.aliases_scope = aliases_scope

    # -- REGISTRATION CORE FUNCTIONALITY:
    def _register_filter(self, func, name=None, aliases=None):
        self._lib.add_filter(func, name, aliases, self.aliases_scope)
        return func

    def _register_test(self, func, name=None, aliases=None):
        self._lib.add_test(func, name, aliases, self.aliases_scope)
        return func

    def _register_global(self, func, name=None, aliases=None):
        self._lib.add_global(func, name, aliases, self.aliases_scope)
        return func

    # -- DECORATOR FUNCTIONS:
    def filter(self, func=None, name=None, aliases=None):
        """Decorator to register a Jinja2 filter (function) to the library.
        If no name is provided, the function-name is used as name.
        Note that aliases (sequence-of-names) allow to register the function
        with additional names.

        :param func:  Function to use as filter function.
        :param name:  Use function-name if None (as string).
        :param aliases: Optional, use for additional names (list-of-strings)
        :return: Decorator or registered function.
        """
        # self._lib.add_filter(name, func)
        if isinstance(func, six.string_types) and name is None:
            # -- CORRECT LAZINESS: @register.filter("name")
            name = func
            func = None

        if func is None:
            # CASE: @register.filter()
            # CASE: @register.filter("name", aliases=...)
            # CASE: @register.filter(name="name", aliases=...)
            def decorator(func):
                return self._register_filter(func, name, aliases)
            return decorator
        elif callable(func):
            # CASE: @register.filter
            return self._register_filter(func, name, aliases)
        else:
            message = "Unsupported args: func=%r, name=%r" % (func, name)
            raise AttributeError(message)

    def test(self, func=None, name=None, aliases=None):
        """Decorator to register a Jinja2 test (function) to the library."""
        if isinstance(func, six.string_types) and name is None:
            # -- CORRECT LAZINESS: @register.test("name")
            name = func
            func = None

        if func is None:
            # CASE: @register.test()
            # CASE: @register.test("name", aliases=...)
            # CASE: @register.test(name="name", aliases=...)
            def decorator(func):
                return self._register_test(func, name, aliases)
            return decorator
        elif callable(func):
            # CASE: @register.test
            return self._register_test(func, name, aliases)
        else:
            message = "Unsupported args: func=%r, name=%r" % (func, name)
            raise AttributeError(message)

    def global_(self, func=None, name=None, aliases=None):
        """Decorator to register a Jinja2 global (function) to the library."""
        if isinstance(func, six.string_types) and name is None:
            # CORRECT LAZINESS: @register.global_("name")
            name = func
            func = None

        if func is None:
            # CASE: @register.global_()
            # CASE: @register.global_("name", aliases=...)
            # CASE: @register.global_(name="name", aliases=...)
            def decorator(func):
                return self._register_global(func, name, aliases)
            return decorator
        elif callable(func):
            # CASE: @register.global_
            return self._register_global(func, name, aliases)
        else:
            message = "Unsupported args: func=%r, name=%r" % (func, name)
            raise AttributeError(message)

    def extension(self, extension_class):
        """Decorator to register a Jinja2 extension class (tag) to the library."""
        assert isinstance(extension_class, six.string_types) or \
               issubclass(extension_class, jinja2.ext.Extension), type(extension_class)
        assert extension_class not in self._lib.extensions
        self._lib.add_extension(extension_class)
        return extension_class



from __future__ import absolute_import, print_function

import jinja2
from jinja2 import Environment
from jinja2_library import add_library_to_environment, Library
import pytest


# ------------------------------------------------------------------------------
# TEST SUPPORT:
# ------------------------------------------------------------------------------
JINJA2_EXTENSIONS = [

]

def make_environment(extensions=None, **kwargs):
    extensions = extensions or JINJA2_EXTENSIONS
    return Environment(extensions=extensions, **kwargs)


def render_template(template_text, environment=None, **data):
    environment = Environment(extensions=JINJA2_EXTENSIONS)
    return environment.from_string(template_text).render(**data)


# ------------------------------------------------------------------------------
# TEST SUITE:
# ------------------------------------------------------------------------------
class TestLibraryUsage(object):
    """Ensure that the library functionality can be used in templates."""

    @staticmethod
    def make_template_with_library(template_text, library):
        """Helper function to simplify tests."""
        environment = make_environment()
        add_library_to_environment(library, environment)
        template = environment.from_string(template_text)
        return template


    def test_use_filter(self):
        # -- LIBRARY MODULE BEGIN -----------------------------------
        library = Library()
        register = library.make_register_decorator()

        @register.filter
        def hello(value, greeting=None):
            greeting = greeting or "Hello"
            return "{greeting} {name}".format(greeting=greeting, name=value)
        # -- LIBRARY MODULE END -------------------------------------

        # -- STEP: Use template with library
        template_text = """{{"Alice"|hello}}"""
        template = self.make_template_with_library(template_text, library)
        output = template.render(name="Alice")
        assert output == "Hello Alice"

    def test_use_filter__with_call_args(self):
        # -- LIBRARY MODULE BEGIN -----------------------------------
        library = Library()
        register = library.make_register_decorator()

        @register.filter
        def hello(value, greeting=None):
            greeting = greeting or "Hello"
            return "{greeting} {name}".format(greeting=greeting, name=value)
        # -- LIBRARY MODULE END -------------------------------------

        # -- STEP: Use template with library
        template_text = """{{"Bob"|hello(greeting="Ciao")}}"""
        template = self.make_template_with_library(template_text, library)
        output = template.render(name="Bob")
        assert output == "Ciao Bob"

    @pytest.mark.parametrize("scope", ["foo", "foo.bar"])
    def test_use_filter__with_scoped_library(self, scope):
        # -- LIBRARY MODULE BEGIN -----------------------------------
        library = Library(scope)
        register = library.make_register_decorator()

        @register.filter
        def hello(value, greeting=None):
            greeting = greeting or "Hello"
            return "{greeting} {name}".format(greeting=greeting, name=value)
        # -- LIBRARY MODULE END -------------------------------------

        # -- STEP: Use template with library
        template_text = """{{"Charly"|%s.hello}}""" % scope
        template = self.make_template_with_library(template_text, library)
        output = template.render(name="Charly")
        assert output == "Hello Charly"
        assert scope in template_text

    @pytest.mark.skipif(not hasattr(jinja2, "pass_context"), reason="FOR: jinja2 >= 3.0")
    def test_use_filter__with_jinja2__pass_context_decorator(self):
        """Ensure that jinja2 filter decorators can be used."""
        # -- LIBRARY MODULE BEGIN -----------------------------------
        library = Library()
        register = library.make_register_decorator()

        @register.filter
        @jinja2.pass_context
        def hello(context, value, greeting=None):
            greeting = greeting or "Hello"
            return "{greeting} {name}".format(greeting=greeting, name=value)
        # -- LIBRARY MODULE END -------------------------------------

        # -- STEP: Use template with library
        template_text = """{{"Dominik"|hello(greeting="Bonjour")}}"""
        template = self.make_template_with_library(template_text, library)
        output = template.render(name="Dominik")
        assert output == "Bonjour Dominik"

    @pytest.mark.skipif(not hasattr(jinja2, "contextfilter"), reason="FOR: jinja2 < 3.0")
    def test_use_filter__with_jinja2__pass_context_decorator(self):
        """Ensure that jinja2 filter decorators can be used."""
        # -- LIBRARY MODULE BEGIN -----------------------------------
        library = Library()
        register = library.make_register_decorator()

        @register.filter
        @jinja2.contextfilter
        def hello(context, value, greeting=None):
            greeting = greeting or "Hello"
            return "{greeting} {name}".format(greeting=greeting, name=value)
        # -- LIBRARY MODULE END -------------------------------------

        # -- STEP: Use template with library
        template_text = """{{"Dominik"|hello(greeting="Bonjour")}}"""
        template = self.make_template_with_library(template_text, library)
        output = template.render(name="Dominik")
        assert output == "Bonjour Dominik"

    def test_use_test(self):
        # -- LIBRARY MODULE BEGIN -----------------------------------
        library = Library()
        register = library.make_register_decorator()

        @register.test
        def even(value):
            return (value % 2) == 0
        # -- LIBRARY MODULE END -------------------------------------

        # -- STEP: Use template with library
        template_text = """TEST: {{ 10 is even }}"""
        template = self.make_template_with_library(template_text, library)
        output = template.render()
        assert output == "TEST: True"

    def test_use_test__with_call_args(self):
        # -- LIBRARY MODULE BEGIN -----------------------------------
        library = Library()
        register = library.make_register_decorator()

        @register.test
        def even2(value1, value2):
            return ((value1 + value2) % 2) == 0
        # -- LIBRARY MODULE END -------------------------------------

        # -- STEP: Use template with library
        template_text = """TEST2: {{ 11 is even2(2) }}"""
        template = self.make_template_with_library(template_text, library)
        output = template.render()
        assert output == "TEST2: False"

    @pytest.mark.parametrize("scope", ["foo", "foo.baz"])
    def test_use_test__with_scoped_library_raises_error(self, scope):
        # -- LIBRARY MODULE BEGIN -----------------------------------
        library = Library(scope)
        register = library.make_register_decorator()

        @register.test
        def even3(value):
            return (value % 2) == 0
        # -- LIBRARY MODULE END

        # -- STEP: Use template with library
        template_text = """TEST3: {{ 10 is %s.even3 }}""" % scope
        template = self.make_template_with_library(template_text, library)
        output = template.render()
        assert output == "TEST3: True"
        assert scope in template_text

    def test_use_global(self):
        # -- LIBRARY MODULE BEGIN -----------------------------------
        library = Library()
        register = library.make_register_decorator()

        @register.global_
        def hello(value, greeting=None):
            greeting = greeting or "Hello"
            return "{greeting} {name}".format(greeting=greeting, name=value)
        # -- LIBRARY MODULE END -------------------------------------

        # -- STEP: Use template with library
        template_text = """TEST: {{ hello(name) }}"""
        template = self.make_template_with_library(template_text, library)
        output = template.render(name="Alice")
        assert output == "TEST: Hello Alice"

    def test_use_global__with_call_args(self):
        # -- LIBRARY MODULE BEGIN -----------------------------------
        library = Library()
        register = library.make_register_decorator()

        @register.global_
        def hello(value, greeting=None):
            greeting = greeting or "Hello"
            return "{greeting} {name}".format(greeting=greeting, name=value)
        # -- LIBRARY MODULE END -------------------------------------

        # -- STEP: Use template with library
        template_text = """TEST: {{ hello(name, greeting="Bonjour") }}"""
        template = self.make_template_with_library(template_text, library)
        output = template.render(name="Bob")
        assert output == "TEST: Bonjour Bob"

    @pytest.mark.parametrize("scope", ["foo", "foo.baz"])
    def test_use_global__with_scoped_library_raises_error(self, scope):
        # -- LIBRARY MODULE BEGIN -----------------------------------
        library = Library(scope)
        register = library.make_register_decorator()

        @register.global_
        def hello(value, greeting=None):
            greeting = greeting or "Hello"
            return "{greeting} {name}".format(greeting=greeting, name=value)
        # -- LIBRARY MODULE END -------------------------------------

        # -- STEP: Use template with library
        template_text = """TEST: {{ %s.hello(name) }}""" % scope
        template = self.make_template_with_library(template_text, library)
        output = template.render(name="Charly")
        assert output == "TEST: Hello Charly"
        assert scope in template_text

    def test_use_extension(self):
        # -- LIBRARY MODULE BEGIN -----------------------------------
        library = Library()
        register = library.make_register_decorator()

        @register.extension
        class HelloExtension(jinja2.ext.Extension):
            tags = set(["hello"])

            def parse(self, parser):
                lineno = next(parser.stream).lineno
                body = parser.parse_statements(["name:endhello"],
                                               drop_needle=True)
                return jinja2.nodes.CallBlock(
                    self.call_method("_hello"), [], [], body).set_lineno(lineno)
                # -- RETURN: nodes.CallBlock/Call that is executed later-on.

            def _hello(self, caller):
                captured_text = caller()
                return "HELLO {0}".format(captured_text)
        # -- LIBRARY MODULE END -------------------------------------

        # -- STEP: Use template with library
        template_text = """TEST: {% hello -%} {{ name2 }} {%- endhello %}"""
        template = self.make_template_with_library(template_text, library)
        output = template.render(name2="Alice")
        assert output == "TEST: HELLO Alice"

    def test_use_extension__registered_by_name(self):
        # -- LIBRARY MODULE BEGIN -----------------------------------
        library = Library()
        library.add_extension("example.hello2.LibraryExtension")
        # -- LIBRARY MODULE END -------------------------------------

        # -- STEP: Use template with library
        this_template = """TEST: {% hello -%} {{ name2 }} {%- endhello %}"""
        template = self.make_template_with_library(this_template, library)
        output = template.render(name2="Bob")
        assert output == "TEST: HELLO Bob"



# -*- coding: utf-8 -*-
"""
Test module for unit-tests of the template library module.

REQUIRE: pytest
"""

from __future__ import absolute_import, print_function
from jinja2 import Environment
from jinja2_library import Library
import pytest
import jinja2
import jinja2.ext


# ------------------------------------------------------------------------------
# TEST SUPPORT:
# ------------------------------------------------------------------------------
JINJA2_EXTENSIONS = [

]

def make_environment(extensions=None, **kwargs):
    extensions = extensions or JINJA2_EXTENSIONS
    return Environment(extensions=extensions, **kwargs)


def render_template(template_text, environment=None, **data):
    environment = Environment(extensions=JINJA2_EXTENSIONS)
    return environment.from_string(template_text).render(**data)


# ------------------------------------------------------------------------------
# TESTSUITE:
# ------------------------------------------------------------------------------
class TestRegisterToLibraryDecorator(object):
    """Test library registration decorator functions."""

    # -- CASE: @register.filter
    def test_filter__as_decorator(self):
        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()

        @register.filter
        def func_filter1(value, **kwargs):
            pass

        # -- CHECK:
        # NOTE: function.__name__ is used name if none is provided.
        registered_func = library.filters.get("func_filter1")
        assert registered_func is func_filter1
        assert len(library.filters) == 1

    def test_filter__as_decorator_call_without_params(self):
        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()

        @register.filter()
        def func_filter2(value, **kwargs):
            pass

        # -- CHECK:
        # NOTE: function.__name__ is used name if none is provided.
        registered_func = library.filters.get("func_filter2")
        assert registered_func is func_filter2
        assert len(library.filters) == 1

    def test_filter__as_decorator_call_with_name_key(self):
        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()

        @register.filter(name="func_filter3_name")
        def func_filter3(value, **kwargs):
            pass

        # -- CHECK:
        registered_func = library.filters.get("func_filter3_name")
        assert registered_func is func_filter3
        assert len(library.filters) == 1

    def test_filter__as_decorator_call_with_name_arg(self):
        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()

        @register.filter("func_filter4_name")
        def func_filter4(value, **kwargs):
            pass

        # -- CHECK:
        registered_func = library.filters.get("func_filter4_name")
        assert "func_filter4_name" in library.filters
        assert registered_func is func_filter4
        assert len(library.filters) == 1

    def test_filter__as_decorator_call_with_aliases(self):
        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()

        @register.filter(aliases=["alias.filter5_1", "alias.filter5_2"])
        def func_filter5(value, **kwargs):
            pass

        # -- CHECK:
        assert set(library.filters.keys()) == \
               set(["func_filter5", "alias.filter5_1", "alias.filter5_2"])
        assert list(library.filters.values()) == ([func_filter5] * 3)
        assert len(library.filters) == 3

    def test_filter__can_call_registered_function(self):
        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()

        @register.filter
        def func_hello(value):
            return "Hello %s" % value

        # -- CHECK:
        registered_func = library.filters.get("func_hello")
        assert registered_func is func_hello
        assert "Hello world" == registered_func("world")

    @pytest.mark.parametrize("name, aliases, error", [
        (123,       None,    AttributeError),  # CASE: name has invalid type.
        ("alice",   ["bob"], AttributeError),  # CASE: Unnamed aliases param.
    ])
    def test_filter__with_invalid_args_raises_error(self, name, aliases, error):
        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()

        # -- NOTE: Unnamed decorator params may cause problems.
        # error_class = error
        with pytest.raises(error):
            @register.filter(name, aliases)
            def func_filter6(value, **kwargs):
                pass

        # -- CHECK:
        assert len(library.filters) == 0


    # -- CASE: @register.test
    def test_test__as_decorator(self):
        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()

        @register.test
        def func_test1(value, **kwargs):
            return True

        # -- CHECK:
        # NOTE: function.__name__ is used name if none is provided.
        registered_func = library.tests.get("func_test1")
        assert registered_func is func_test1
        assert len(library.tests) == 1

    def test_test__as_decorator_call_without_params(self):
        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()

        @register.test()
        def func_test2(value, **kwargs):
            return True

        # -- CHECK:
        # NOTE: function.__name__ is used name if none is provided.
        registered_func = library.tests.get("func_test2")
        assert registered_func is func_test2
        assert len(library.tests) == 1

    def test_test__as_decorator_call_with_name_key(self):
        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()

        @register.test(name="func_test3_name")
        def func_test3(value, **kwargs):
            return True

        # -- CHECK:
        registered_func = library.tests.get("func_test3_name")
        assert registered_func is func_test3
        assert len(library.tests) == 1

    def test_test__as_decorator_call_with_name_arg(self):
        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()

        @register.test("func_test4_name")
        def func_test4(value, **kwargs):
            return True

        # -- CHECK:
        registered_func = library.tests.get("func_test4_name")
        assert "func_test4_name" in library.tests
        assert registered_func is func_test4
        assert len(library.tests) == 1

    def test_test__as_decorator_call_with_aliases(self):
        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()

        @register.test(aliases=["alias.test5_1", "alias.test5_2"])
        def func_test5(value, **kwargs):
            return True

        # -- CHECK:
        assert set(library.tests.keys()) == \
               set(["func_test5", "alias.test5_1", "alias.test5_2"])
        assert list(library.tests.values()) == ([func_test5] * 3)
        assert len(library.tests) == 3

    def test_test__can_call_registered_function(self):
        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()

        @register.test
        def func_true(value):
            return bool(value)

        # -- CHECK:
        registered_func = library.tests.get("func_true")
        assert registered_func is func_true
        assert True  == registered_func(1)
        assert False == registered_func(0)

    @pytest.mark.parametrize("name, aliases, error", [
        (123,       None,    AttributeError),  # CASE: name has invalid type.
        ("alice",   ["bob"], AttributeError),  # CASE: Unnamed aliases param.
    ])
    def test_test__with_invalid_args_raises_error(self, name, aliases, error):
        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()

        # -- NOTE: Unnamed decorator params may cause problems.
        # error_class = error
        with pytest.raises(error):
            @register.test(name, aliases)
            def func_test6(value):
                return False

        # -- CHECK:
        assert len(library.tests) == 0


    # -- CASE: @register.global_
    def test_global__as_decorator(self):
        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()

        @register.global_
        def func_global1(value, **kwargs):
            pass

        # -- CHECK:
        # NOTE: function.__name__ is used name if none is provided.
        registered_func = library.globals.get("func_global1")
        assert registered_func is func_global1
        assert len(library.globals) == 1

    def test_global__as_decorator_call_without_params(self):
        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()

        @register.global_()
        def func_global2(value, **kwargs):
            pass

        # -- CHECK:
        # NOTE: function.__name__ is used name if none is provided.
        registered_func = library.globals.get("func_global2")
        assert registered_func is func_global2
        assert len(library.globals) == 1

    def test_global__as_decorator_call_with_name_key(self):
        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()

        @register.global_(name="func_global3_name")
        def func_global3(value, **kwargs):
            pass

        # -- CHECK:
        registered_func = library.globals.get("func_global3_name")
        assert registered_func is func_global3
        assert len(library.globals) == 1

    def test_global__as_decorator_call_with_name_arg(self):
        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()

        @register.global_("func_global4_name")
        def func_global4(value, **kwargs):
            pass

        # -- CHECK:
        registered_func = library.globals.get("func_global4_name")
        assert "func_global4_name" in library.globals
        assert registered_func is func_global4
        assert len(library.globals) == 1

    def test_global__as_decorator_call_with_aliases(self):
        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()

        @register.global_(aliases=["alias.global5_1", "alias.global5_2"])
        def func_global5(value, **kwargs):
            pass

        # -- CHECK:
        assert set(library.globals.keys()) == \
               set(["func_global5", "alias.global5_1", "alias.global5_2"])
        assert list(library.globals.values()) == ([func_global5] * 3)
        assert len(library.globals) == 3

    def test_global__can_call_registered_function(self):
        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()

        @register.global_
        def func_hello(value):
            return "Hello %s" % value

        # -- CHECK:
        registered_func = library.globals.get("func_hello")
        assert registered_func is func_hello
        assert "Hello world" == registered_func("world")

    @pytest.mark.parametrize("name, aliases, error", [
        (123,       None,    AttributeError),  # CASE: name has invalid type.
        ("alice",   ["bob"], AttributeError),  # CASE: Unnamed aliases param.
    ])
    def test_global__with_invalid_args_raises_error(self, name, aliases, error):
        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()

        # -- NOTE: Unnamed decorator params may cause problems.
        # error_class = error
        with pytest.raises(error):
            @register.global_(name, aliases)
            def func_global6(value, **kwargs):
                pass

        # -- CHECK:
        assert len(library.globals) == 0

    def test_extension__with_extension_class(self):

        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()

        @register.extension
        class AliceExtension(jinja2.ext.Extension):
            tags = set(["alice"])
            def parse(self, parser): pass

        # -- CHECK:
        assert library.extensions == [AliceExtension]

    def test_extension__raises_error_for_known_extension(self):

        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()

        @register.extension
        class AliceExtension(jinja2.ext.Extension):
            tags = set(["alice"])
            def parse(self, parser): pass

        # -- CHECK: Register extension again.
        with pytest.raises(AssertionError):
            register.extension(AliceExtension)


    def test_extension__with_extension_name(self):

        # -- BLUEPRINT:
        library = Library()
        register = library.make_register_decorator()
        register.extension("my_module.AliceExtension")

        # -- CHECK:
        assert library.extensions == ["my_module.AliceExtension"]

class TestLibrary(object):

    def test_add_filter__with_unscoped_library(self):
        library = Library()
        def my_filter(x):
            pass

        library.add_filter(my_filter, "my_filter")
        assert "my_filter" in library.filters
        assert library.filters["my_filter"] is my_filter

    @pytest.mark.parametrize("scope", ["foo", "foo.bar"])
    def test_add_filter__with_scoped_library(self, scope):
        library = Library(scope)
        def my_filter(x):
            pass

        unscoped_name = "my_filter"
        scoped_name = "%s.%s" % (scope, unscoped_name)
        library.add_filter(my_filter, "my_filter")
        assert scoped_name in library.filters
        assert unscoped_name not in library.filters
        assert library.filters[scoped_name] is my_filter


    def test_add_test__with_unscoped_library(self):
        library = Library()
        def my_test(value):
            return bool(value)

        library.add_test(my_test, "my_test")
        assert "my_test" in library.tests
        assert library.tests["my_test"] is my_test

    @pytest.mark.parametrize("scope", ["foo", "foo.bar"])
    def test_add_test__with_scoped_library(self, scope):
        library = Library(scope)
        def my_test(value):
            return bool(value)

        library.add_test(my_test, "my_test")
        unscoped_name = "my_test"
        scoped_name = "%s.%s" % (scope, unscoped_name)
        assert scoped_name in library.tests
        assert unscoped_name not in library.tests
        assert library.tests[scoped_name] is my_test

    def test_add_global__with_unscoped_library(self):
        library = Library()
        def my_global(value):
            return value

        library.add_global(my_global, "my_global")
        assert "my_global" in library.globals
        assert library.globals["my_global"] is my_global

    @pytest.mark.parametrize("scope", ["foo", "foo.bar"])
    def test_add_global__with_scoped_library(self, scope):
        library = Library(scope)
        def my_global(value):
            return value

        library.add_global(my_global, "my_global")

        unscoped_name = "my_global"
        scoped_name = "%s.%s" % (scope, unscoped_name)
        assert scoped_name in library.globals
        assert unscoped_name not in library.globals
        assert library.globals[scoped_name] is my_global

    def test_add_extension__by_class(self):
        import jinja2.ext
        class MyExtension(jinja2.ext.Extension): pass

        library = Library()
        library.add_extension(MyExtension)
        assert MyExtension in library.extensions

    def test_add_extension__by_scoped_name(self):
        library = Library()
        library.add_extension("jinja2.ext.i18n")
        assert "jinja2.ext.i18n" in library.extensions



from __future__ import absolute_import, print_function
import os.path
import sys
import pytest

try:
    # -- CASE: Python >= 3.5
    from subprocess import run
except ImportError:
    # -- CASE: Python < 3.5
    run = None

# -----------------------------------------------------------------------------
# CONSTANTS
# -----------------------------------------------------------------------------
HERE = os.path.dirname(__file__)
TOPDIR = os.path.abspath(os.path.dirname(HERE) or "..")
PYTHON = sys.executable

_ON_WINDOWS = sys.platform.startswith("win32")
_PYTHON_VERSION = sys.version_info[:2]
_CI_ON_WINDOWS_WITH_PY310 = (_ON_WINDOWS
    and ("CI" in os.environ)
    and (_PYTHON_VERSION == (3, 10))
)

# -----------------------------------------------------------------------------
# TEST SUITE
# -----------------------------------------------------------------------------
@pytest.mark.skipif(run is None, reason="MISSING: subprocess.run()")
@pytest.mark.skipif(_CI_ON_WINDOWS_WITH_PY310, reason="CI: Python Windows platform problem")
class TestLibraryExamples:
    EXAMPLES_EXPECTED_SCRIPT = [
        ("HELLO: Ciao Alice", "example/use_template_library_hello.py"),
        ("HELLO: Ciao Bob",   "example/use_template_library_scoped_hello.py"),
        ("HELLO_2: Ciao Charly", "example/use_template_library_hello2.py"),
    ]

    @pytest.mark.parametrize("expected, script", EXAMPLES_EXPECTED_SCRIPT)
    def test_example(self, expected, script):
        print("SCRIPT: {}".format(script))
        command = "{python} {script}".format(python=PYTHON, script=script)
        result = run(command, capture_output=True, shell=True,
                     env=dict(PYTHONPATH=TOPDIR))
        captured_output = result.stdout.decode("utf-8")
        print(captured_output)
        if result.stderr:
            captured_stderr = result.stderr.decode("utf-8").strip()
            print(captured_stderr, file=sys.stderr)
        assert expected in captured_output
        assert "AssertionError" not in captured_output
        assert result.returncode == 0
